import os
import urllib.parse
from typing import Dict, List, Optional, Tuple

import aiohttp
from chainlit.client.base import AppUser
from fastapi import HTTPException


class OAuthProvider:
    id: str
    env: List[str]
    client_id: str
    client_secret: str
    authorize_url: str
    authorize_params: Dict[str, str]

    def is_configured(self):
        return all([os.environ.get(env) for env in self.env])

    async def get_token(self, code: str, url: str) -> str:
        raise NotImplementedError()

    async def get_user_info(self, token: str) -> Tuple[Dict[str, str], AppUser]:
        raise NotImplementedError()


class GithubOAuthProvider(OAuthProvider):
    id = "github"
    env = ["OAUTH_GITHUB_CLIENT_ID", "OAUTH_GITHUB_CLIENT_SECRET"]
    authorize_url = "https://github.com/login/oauth/authorize"

    def __init__(self):
        self.client_id = os.environ.get("OAUTH_GITHUB_CLIENT_ID")
        self.client_secret = os.environ.get("OAUTH_GITHUB_CLIENT_SECRET")
        self.authorize_params = {
            "scope": "user:email",
        }

    async def get_token(self, code: str, url: str):
        payload = {
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "code": code,
        }
        async with aiohttp.ClientSession(raise_for_status=True) as session:
            async with session.post(
                "https://github.com/login/oauth/access_token",
                json=payload,
            ) as result:
                text = await result.text()
                content = urllib.parse.parse_qs(text)
                token = content.get("access_token", [""])[0]
                if not token:
                    raise HTTPException(
                        status_code=400, detail="Failed to get the access token"
                    )
                return token

    async def get_user_info(self, token: str):
        async with aiohttp.ClientSession(raise_for_status=True) as session:
            async with session.get(
                "https://api.github.com/user",
                headers={"Authorization": f"token {token}"},
            ) as result:
                user = await result.json()

                async with session.get(
                    "https://api.github.com/user/emails",
                    headers={"Authorization": f"token {token}"},
                ) as email_result:
                    emails = await email_result.json()

                    user.update({"emails": emails})

                    app_user = AppUser(
                        username=user["login"],
                        image=user["avatar_url"],
                        provider="github",
                    )
                    return (user, app_user)


class GoogleOAuthProvider(OAuthProvider):
    id = "google"
    env = ["OAUTH_GOOGLE_CLIENT_ID", "OAUTH_GOOGLE_CLIENT_SECRET"]
    authorize_url = "https://accounts.google.com/o/oauth2/v2/auth"

    def __init__(self):
        self.client_id = os.environ.get("OAUTH_GOOGLE_CLIENT_ID")
        self.client_secret = os.environ.get("OAUTH_GOOGLE_CLIENT_SECRET")
        self.authorize_params = {
            "scope": "https://www.googleapis.com/auth/userinfo.profile https://www.googleapis.com/auth/userinfo.email",
            "response_type": "code",
            "access_type": "offline",
        }

    async def get_token(self, code: str, url: str):
        payload = {
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "code": code,
            "grant_type": "authorization_code",
            "redirect_uri": url,
        }
        async with aiohttp.ClientSession(raise_for_status=True) as session:
            async with session.post(
                "https://oauth2.googleapis.com/token",
                data=payload,
            ) as result:
                json = await result.json()
                token = json["access_token"]
                if not token:
                    raise HTTPException(
                        status_code=400, detail="Failed to get the access token"
                    )
                return token

    async def get_user_info(self, token: str):
        async with aiohttp.ClientSession(raise_for_status=True) as session:
            async with session.get(
                "https://www.googleapis.com/userinfo/v2/me",
                headers={"Authorization": f"Bearer {token}"},
            ) as result:
                user = await result.json()

                app_user = AppUser(
                    username=user["name"], image=user["picture"], provider="google"
                )
                return (user, app_user)


class AzureADOAuthProvider(OAuthProvider):
    id = "azure-ad"
    env = [
        "OAUTH_AZURE_AD_CLIENT_ID",
        "OAUTH_AZURE_AD_CLIENT_SECRET",
        "OAUTH_AZURE_AD_TENANT_ID",
    ]
    authorize_url = "https://login.microsoftonline.com/common/oauth2/v2.0/authorize"

    def __init__(self):
        self.client_id = os.environ.get("OAUTH_AZURE_AD_CLIENT_ID")
        self.client_secret = os.environ.get("OAUTH_AZURE_AD_CLIENT_SECRET")
        self.authorize_params = {
            "tenant": os.environ.get("OAUTH_AZURE_AD_TENANT_ID"),
            "response_type": "code",
            "scope": "https://graph.microsoft.com/User.Read",
            "response_mode": "query",
        }

    async def get_token(self, code: str, url: str):
        payload = {
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "code": code,
            "grant_type": "authorization_code",
            "redirect_uri": url,
        }
        async with aiohttp.ClientSession(raise_for_status=True) as session:
            async with session.post(
                "https://login.microsoftonline.com/common/oauth2/v2.0/token",
                data=payload,
            ) as result:
                json = await result.json()

                token = json["access_token"]
                if not token:
                    raise HTTPException(
                        status_code=400, detail="Failed to get the access token"
                    )
                return token

    async def get_user_info(self, token: str):
        async with aiohttp.ClientSession(raise_for_status=True) as session:
            async with session.get(
                "https://graph.microsoft.com/v1.0/users/me",
                headers={"Authorization": f"Bearer {token}"},
            ) as result:
                user = await result.json()

                app_user = AppUser(
                    username=user["userPrincipalName"], image="", provider="azure-ad"
                )
                return (user, app_user)


class OktaOAuthProvider(OAuthProvider):
    id = "okta"
    env = [
        "OAUTH_OKTA_CLIENT_ID",
        "OAUTH_OKTA_CLIENT_SECRET",
        "OAUTH_OKTA_DOMAIN",
    ]
    # Avoid trailing slash in domain if supplied
    domain = f"https://{os.environ.get('OAUTH_OKTA_DOMAIN', '').rstrip('/')}"

    authorize_url = f"{domain}/oauth2/default/v1/authorize"

    def __init__(self):
        self.client_id = os.environ.get("OAUTH_OKTA_CLIENT_ID")
        self.client_secret = os.environ.get("OAUTH_OKTA_CLIENT_SECRET")
        self.authorize_params = {
            "response_type": "code",
            "scope": "openid profile email",
            "response_mode": "query",
        }

    async def get_token(self, code: str, url: str):
        payload = {
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "code": code,
            "grant_type": "authorization_code",
            "redirect_uri": url,
        }
        async with aiohttp.ClientSession(raise_for_status=True) as session:
            async with session.post(
                f"{self.domain}/oauth2/default/v1/token",
                data=payload,
            ) as result:
                json = await result.json()

                token = json["access_token"]
                if not token:
                    raise HTTPException(
                        status_code=400, detail="Failed to get the access token"
                    )
                return token

    async def get_user_info(self, token: str):
        async with aiohttp.ClientSession(raise_for_status=True) as session:
            async with session.get(
                f"{self.domain}/oauth2/default/v1/userinfo",
                headers={"Authorization": f"Bearer {token}"},
            ) as result:
                user = await result.json()

                app_user = AppUser(
                    username=user.get("email"), image="", provider="okta"
                )
                return (user, app_user)


providers = [GithubOAuthProvider(), GoogleOAuthProvider(), AzureADOAuthProvider(), OktaOAuthProvider()]


def get_oauth_provider(provider: str) -> Optional[OAuthProvider]:
    for p in providers:
        if p.id == provider:
            return p
    return None


def get_configured_oauth_providers():
    return [p.id for p in providers if p.is_configured()]
                                                                                                                                                                                                                                                                                                                                                                        



from typing import TYPE_CHECKING, Dict, Optional, TypedDict, Union

if TYPE_CHECKING:
    from chainlit.message import Message
    from chainlit.client.base import AppUser, PersistedAppUser

from chainlit.context import context


class UserSessionDict(TypedDict):
    id: str
    env: Dict[str, str]
    headers: Dict[str, str]
    user: Optional[Union["AppUser", "PersistedAppUser"]]
    root_message: Optional["Message"]


user_sessions: Dict[str, UserSessionDict] = {}


class UserSession:
    """
    Developer facing user session class.
    Useful for the developer to store user specific data between calls.
    """

    def get(self, key, default=None):
        if not context.emitter:
            return default

        if context.session.id not in user_sessions:
            # Create a new user session
            user_sessions[context.session.id] = {}

        user_session = user_sessions[context.session.id]

        # Copy important fields from the session
        user_session["id"] = context.session.id
        user_session["env"] = context.session.user_env
        user_session["chat_settings"] = context.session.chat_settings
        user_session["user"] = context.session.user

        if context.session.root_message:
            user_session["root_message"] = context.session.root_message

        return user_session.get(key, default)

    def set(self, key, value):
        if not context.emitter:
            return None

        if context.session.id not in user_sessions:
            user_sessions[context.session.id] = {}

        user_session = user_sessions[context.session.id]
        user_session[key] = value


user_session = UserSession()



import asyncio
from contextvars import ContextVar
from typing import TYPE_CHECKING

from chainlit.session import Session
from lazify import LazyProxy

if TYPE_CHECKING:
    from chainlit.emitter import ChainlitEmitter


class ChainlitContextException(Exception):
    def __init__(self, msg="Chainlit context not found", *args, **kwargs):
        super().__init__(msg, *args, **kwargs)


class ChainlitContext:
    loop: asyncio.AbstractEventLoop
    emitter: "ChainlitEmitter"
    session: Session

    def __init__(self, session: Session):
        from chainlit.emitter import ChainlitEmitter

        self.loop = asyncio.get_running_loop()
        self.session = session
        self.emitter = ChainlitEmitter(session)


context_var: ContextVar[ChainlitContext] = ContextVar("chainlit")


def init_context(session_or_sid) -> ChainlitContext:
    if not isinstance(session_or_sid, Session):
        session_or_sid = Session.require(session_or_sid)

    context = ChainlitContext(session_or_sid)
    context_var.set(context)
    return context


def get_context() -> ChainlitContext:
    try:
        return context_var.get()
    except LookupError:
        raise ChainlitContextException()


context: ChainlitContext = LazyProxy(get_context, enable_cache=False)



from enum import Enum
from typing import Dict, List, Literal, TypedDict, Union

from chainlit.client.base import ConversationFilter, Pagination
from chainlit.prompt import Prompt
from dataclasses_json import DataClassJsonMixin
from pydantic import BaseModel
from pydantic.dataclasses import dataclass

InputWidgetType = Literal[
    "switch", "slider", "select", "textinput", "tags", "numberinput"
]


@dataclass
class FileSpec(DataClassJsonMixin):
    accept: Union[List[str], Dict[str, List[str]]]
    max_files: int
    max_size_mb: int


@dataclass
class AskSpec(DataClassJsonMixin):
    """Specification for asking the user."""

    timeout: int
    type: Literal["text", "file"]


@dataclass
class AskFileSpec(FileSpec, AskSpec, DataClassJsonMixin):
    """Specification for asking the user a file."""


class AskResponse(TypedDict):
    content: str
    author: str


@dataclass
class AskFileResponse:
    name: str
    path: str
    size: int
    type: str
    content: bytes


class CompletionRequest(BaseModel):
    prompt: Prompt
    userEnv: Dict[str, str]


class UpdateFeedbackRequest(BaseModel):
    messageId: str
    feedback: Literal[-1, 0, 1]


class DeleteConversationRequest(BaseModel):
    conversationId: str


class GetConversationsRequest(BaseModel):
    pagination: Pagination
    filter: ConversationFilter


class Theme(str, Enum):
    light = "light"
    dark = "dark"



from chainlit.utils import check_module_version

if not check_module_version("langflow", "0.1.4"):
    raise ValueError(
        "Expected Langflow version >= 0.1.4. Run `pip install langflow --upgrade`"
    )

from typing import Dict, Optional, Union

import aiohttp
from chainlit.telemetry import trace_event


async def load_flow(schema: Union[Dict, str], tweaks: Optional[Dict] = None):
    from langflow import load_flow_from_json

    trace_event("load_langflow")

    if type(schema) == str:
        async with aiohttp.ClientSession() as session:
            async with session.get(
                schema,
            ) as r:
                if not r.ok:
                    reason = await r.text()
                    raise ValueError(f"Error: {reason}")
                schema = await r.json()

    flow = load_flow_from_json(flow=schema, tweaks=tweaks)

    return flow



from abc import ABC, abstractmethod
from collections import defaultdict
from typing import Any, Dict, List, Optional

from chainlit.types import InputWidgetType
from pydantic.dataclasses import Field, dataclass


@dataclass
class InputWidget(ABC):
    id: str
    label: str
    initial: Any = None
    tooltip: Optional[str] = None
    description: Optional[str] = None

    def __post_init__(self, tooltip, description) -> None:
        if not self.id or not self.label:
            raise ValueError("Must provide key and label to load InputWidget")

        self.tooltip = tooltip
        self.description = description

    @abstractmethod
    def to_dict(self) -> Dict[str, Any]:
        pass


@dataclass
class Switch(InputWidget):
    """Useful to create a switch input."""

    type: InputWidgetType = "switch"
    initial: bool = False

    def __init__(
        self,
        id: str,
        label: str,
        initial: bool,
        tooltip: Optional[str] = None,
        description: Optional[str] = None,
    ) -> None:
        self.id = id
        self.label = label
        self.initial = initial
        super().__post_init__(tooltip, description)

    def to_dict(self) -> Dict[str, Any]:
        return {
            "type": self.type,
            "id": self.id,
            "label": self.label,
            "initial": self.initial,
            "tooltip": self.tooltip,
            "description": self.description,
        }


@dataclass
class Slider(InputWidget):
    """Useful to create a slider input."""

    type: InputWidgetType = "slider"
    initial: float = 0
    min: float = 0
    max: float = 10
    step: float = 1

    def __init__(
        self,
        id: str,
        label: str,
        initial: float,
        min: float = 0,
        max: float = 10,
        step: float = 1,
        tooltip: Optional[str] = None,
        description: Optional[str] = None,
    ) -> None:
        self.id = id
        self.label = label
        self.initial = initial
        self.min = min
        self.max = max
        self.step = step
        super().__post_init__(tooltip, description)

    def to_dict(self) -> Dict[str, Any]:
        return {
            "type": self.type,
            "id": self.id,
            "label": self.label,
            "initial": self.initial,
            "min": self.min,
            "max": self.max,
            "step": self.step,
            "tooltip": self.tooltip,
            "description": self.description,
        }


@dataclass
class Select(InputWidget):
    """Useful to create a select input."""

    type: InputWidgetType = "select"
    initial: Optional[str] = None
    initial_index: Optional[int] = None
    initial_value: Optional[str] = None
    values: List[str] = Field(default_factory=lambda: list)
    items: Dict[str, str] = Field(default_factory=lambda: defaultdict(dict))

    def __init__(
        self,
        id: str,
        label: str,
        values: Optional[List[str]] = None,
        items: Optional[Dict[str, str]] = None,
        initial_index: Optional[int] = None,
        initial_value: Optional[str] = None,
        tooltip: Optional[str] = None,
        description: Optional[str] = None,
    ) -> None:
        if values is None and items is None:
            raise ValueError("Must provide values or items to create a Select")

        if values is not None and items is not None:
            raise ValueError(
                "You can only provide either values or items to create a Select"
            )

        if values is None and initial_index is not None:
            raise ValueError(
                "Initial_index can only be used in combination with values to create a Select"
            )

        self.id = id
        self.label = label

        if items is not None:
            self.items = items
            self.initial = initial_value
        elif values is not None:
            self.items = {value: value for value in values}
            self.initial = (
                values[initial_index] if initial_index is not None else initial_value
            )
        super().__post_init__(tooltip, description)

    def to_dict(self) -> Dict[str, Any]:
        return {
            "type": self.type,
            "id": self.id,
            "label": self.label,
            "initial": self.initial,
            "items": [
                {"label": id, "value": value} for id, value in self.items.items()
            ],
            "tooltip": self.tooltip,
            "description": self.description,
        }


@dataclass
class TextInput(InputWidget):
    """Useful to create a text input."""

    type: InputWidgetType = "textinput"
    initial: Optional[str] = None
    placeholder: Optional[str] = None

    def __init__(
        self,
        id: str,
        label: str,
        initial: Optional[str] = None,
        placeholder: Optional[str] = None,
        tooltip: Optional[str] = None,
        description: Optional[str] = None,
    ) -> None:
        self.id = id
        self.label = label
        self.initial = initial
        self.placeholder = placeholder
        super().__post_init__(tooltip, description)

    def to_dict(self) -> Dict[str, Any]:
        return {
            "type": self.type,
            "id": self.id,
            "label": self.label,
            "initial": self.initial,
            "placeholder": self.placeholder,
            "tooltip": self.tooltip,
            "description": self.description,
        }


@dataclass
class NumberInput(InputWidget):
    """Useful to create a number input."""

    type: InputWidgetType = "numberinput"
    initial: Optional[float] = None
    placeholder: Optional[str] = None

    def __init__(
        self,
        id: str,
        label: str,
        initial: Optional[float] = None,
        placeholder: Optional[str] = None,
        tooltip: Optional[str] = None,
        description: Optional[str] = None,
    ) -> None:
        self.id = id
        self.label = label
        self.initial = initial
        self.placeholder = placeholder
        super().__post_init__(tooltip, description)

    def to_dict(self) -> Dict[str, Any]:
        return {
            "type": self.type,
            "id": self.id,
            "label": self.label,
            "initial": self.initial,
            "placeholder": self.placeholder,
            "tooltip": self.tooltip,
            "description": self.description,
        }


@dataclass
class Tags(InputWidget):
    """Useful to create an input for an array of strings."""

    type: InputWidgetType = "tags"
    initial: List[str] = Field(default_factory=lambda: list)
    values: List[str] = Field(default_factory=lambda: list)

    def __init__(
        self,
        id: str,
        label: str,
        initial: Optional[List[str]] = None,
        tooltip: Optional[str] = None,
        description: Optional[str] = None,
    ) -> None:
        self.id = id
        self.label = label
        if initial is not None:
            self.initial = initial
        super().__post_init__(tooltip, description)

    def to_dict(self) -> Dict[str, Any]:
        return {
            "type": self.type,
            "id": self.id,
            "label": self.label,
            "initial": self.initial,
            "tooltip": self.tooltip,
            "description": self.description,
        }



import json
from typing import Any, Dict, List, Optional, Union

from chainlit.config import config
from chainlit.context import context
from chainlit.message import ErrorMessage, Message
from chainlit.prompt import Prompt, PromptMessage
from chainlit.sync import run_sync
from langchain.callbacks.base import AsyncCallbackHandler, BaseCallbackHandler
from langchain.schema import AgentAction, AgentFinish, BaseMessage, LLMResult

IGNORE_LIST = []  # type: List[str]
DEFAULT_ANSWER_PREFIX_TOKENS = ["Final", "Answer", ":"]


def get_llm_settings(invocation_params: Union[Dict, None], serialized: Dict[str, Any]):
    if invocation_params is None:
        return None, None

    provider = invocation_params.pop("_type", "")  # type: str

    if provider.startswith("openai"):
        model_name = invocation_params.pop("model_name")
        invocation_params["model"] = model_name

    model_kwargs = invocation_params.pop("model_kwargs", {})

    merged = {
        **invocation_params,
        **model_kwargs,
        **serialized.get("kwargs", {}),
    }

    # make sure there is no api key specification
    settings = {k: v for k, v in merged.items() if not k.endswith("_api_key")}

    return provider, settings


def build_prompt(serialized: Dict[str, Any], inputs: Dict[str, Any]):
    prompt_params = serialized.get("kwargs", {}).get("prompt", {}).get("kwargs", {})
    _messages = prompt_params.get("messages")

    if _messages:
        messages = []
        for m in _messages:
            class_name = m["id"][-1]
            # A placeholder holds a variable that itself is a list of messages, like chat_history
            if class_name == "MessagesPlaceholder":
                variable_name = m.get("kwargs", {}).get("variable_name")  # type: str
                variable = inputs.get(variable_name, [])
                placeholder_size = len(variable)
                if placeholder_size:
                    messages += [PromptMessage(placeholder_size=placeholder_size)]
            else:
                m_prompt_params = (
                    m.get("kwargs", {}).get("prompt", {}).get("kwargs", {})
                )
                m_template = m_prompt_params.get("template")
                m_template_format = m_prompt_params.get("template_format")
                messages += [
                    PromptMessage(
                        template=m_template,
                        template_format=m_template_format,
                        role=convert_role(class_name),
                    )
                ]
    else:
        messages = None

    template = prompt_params.get("template")
    template_format = prompt_params.get("template_format")

    stringified_inputs = inputs = {k: str(v) for (k, v) in inputs.items()}

    if template:
        return Prompt(
            template=template,
            template_format=template_format,
            inputs=stringified_inputs,
        )
    elif messages:
        return Prompt(inputs=stringified_inputs, messages=messages)


def convert_role(role: str):
    if role in ["human", "chat", "HumanMessage", "HumanMessagePromptTemplate"]:
        return "user"
    elif role in ["system", "SystemMessage", "SystemMessagePromptTemplate"]:
        return "system"
    elif role in ["ai", "AIMessage", "AIMessagePromptTemplate"]:
        return "assistant"
    elif role in ["function", "FunctionMessage", "FunctionMessagePromptTemplate"]:
        return "function"
    else:
        raise ValueError(f"Unsupported role {role}")


def convert_message(message: BaseMessage, template: Optional[str] = None):
    function_call = message.additional_kwargs.get("function_call")
    if function_call:
        content = json.dumps(function_call, indent=4)
    else:
        content = message.content
    return PromptMessage(
        name=getattr(message, "name", None),
        role=convert_role(message.type),
        template=template,
        formatted=content,
    )


class BaseLangchainCallbackHandler(BaseCallbackHandler):
    # Keep track of the prompt sequence
    prompt_sequence: List[Union[Prompt, None]]
    # Keep track of the call sequence, like [AgentExecutor, LLMMathChain, Calculator, ...]
    sequence: List[Message]
    # Keep track of the currently streamed message for the session
    stream: Union[Message, None]
    # The stream we can use to stream the final answer from a chain
    final_stream: Union[Message, None]
    # Message at the root of the chat we should attach child messages to
    root_message: Message
    # Should we stream the final answer?
    stream_final_answer: bool = False
    # Token sequence that prefixes the answer
    answer_prefix_tokens: List[str]
    # Ignore white spaces and new lines when comparing answer_prefix_tokens to last tokens? (to determine if answer has been reached)
    strip_tokens: bool
    # Should answer prefix itself also be streamed?
    stream_prefix: bool

    raise_error = True

    # We want to handler to be called on every message
    always_verbose: bool = True

    def __init__(
        self,
        *,
        answer_prefix_tokens: Optional[List[str]] = None,
        strip_tokens: bool = True,
        stream_prefix: bool = False,
        stream_final_answer: bool = False,
        root_message: Optional[Message] = None,
    ) -> None:
        self.sequence = []
        self.prompt_sequence = []
        self.stream = None

        if root_message:
            self.root_message = root_message
        elif root_message := context.session.root_message:
            self.root_message = root_message
        else:
            self.root_message = Message(author=config.ui.name, content="")
            run_sync(self.root_message.send())

        # Langchain final answer streaming logic
        if answer_prefix_tokens is None:
            self.answer_prefix_tokens = DEFAULT_ANSWER_PREFIX_TOKENS
        else:
            self.answer_prefix_tokens = answer_prefix_tokens
        if strip_tokens:
            self.answer_prefix_tokens_stripped = [
                token.strip() for token in self.answer_prefix_tokens
            ]
        else:
            self.answer_prefix_tokens_stripped = self.answer_prefix_tokens
        self.last_tokens = [""] * len(self.answer_prefix_tokens)
        self.last_tokens_stripped = [""] * len(self.answer_prefix_tokens)
        self.strip_tokens = strip_tokens
        self.stream_prefix = stream_prefix
        self.answer_reached = False

        # Our own final answer streaming logic
        self.stream_final_answer = stream_final_answer
        self.final_stream = None
        self.has_streamed_final_answer = False

    @property
    def current_prompt(self):
        if self.prompt_sequence:
            return self.prompt_sequence[-1]
        else:
            return None

    def append_to_last_tokens(self, token: str) -> None:
        self.last_tokens.append(token)
        self.last_tokens_stripped.append(token.strip())
        if len(self.last_tokens) > len(self.answer_prefix_tokens):
            self.last_tokens.pop(0)
            self.last_tokens_stripped.pop(0)

    def _compare_last_tokens(self, last_tokens: List[str]):
        if last_tokens == self.answer_prefix_tokens_stripped:
            # If tokens match perfectly we are done
            return True
        else:
            # Some LLMs will consider all the tokens of the final answer as one token
            # so we check if any last token contains all answer tokens
            return any(
                [
                    all(
                        answer_token in last_token
                        for answer_token in self.answer_prefix_tokens_stripped
                    )
                    for last_token in last_tokens
                ]
            )

    def check_if_answer_reached(self) -> bool:
        if self.strip_tokens:
            return self._compare_last_tokens(self.last_tokens_stripped)

        else:
            return self._compare_last_tokens(self.last_tokens)

    def start_stream(self):
        author = self.get_author()
        if author in IGNORE_LIST:
            return

        parent_id = self.get_last_message().parent_id

        self.stream = self.create_message(
            prompt=self.current_prompt, author=author, parent_id=parent_id
        )

    def end_stream(self):
        self.stream = None

    def add_in_sequence(self, message: Message):
        self.sequence.append(message)

    def pop_sequence(self):
        if self.sequence:
            return self.sequence.pop()

    def get_author(self):
        if self.sequence:
            return self.sequence[-1].author
        return config.ui.name

    def get_last_message(self):
        for message in reversed(self.sequence):
            if message.author not in IGNORE_LIST:
                return message
        return self.root_message

    def create_error(self, error: Exception):
        if isinstance(error, InterruptedError):
            return None

        return ErrorMessage(content=str(error), author=self.get_author())

    def create_message(
        self,
        content: str = "",
        prompt: Optional[Prompt] = None,
        author: Optional[str] = None,
        parent_id: Optional[str] = None,
    ):
        if parent_id is None:
            last_message = self.get_last_message()
            parent_id = last_message.id

        return Message(
            content,
            author=author or self.get_author(),
            prompt=prompt,
            parent_id=parent_id,
        )


def _on_chat_model_start(
    self: BaseLangchainCallbackHandler,
    serialized: Dict[str, Any],
    messages: List[List[BaseMessage]],
    **kwargs: Any,
):
    invocation_params = kwargs.get("invocation_params")
    provider, settings = get_llm_settings(invocation_params, serialized)

    formatted_messages = messages[0]

    if self.current_prompt:
        self.current_prompt.provider = provider
        self.current_prompt.settings = settings
        # Chat mode
        if self.current_prompt.messages:
            # This is needed to compute the correct message index to read
            placeholder_offset = 0
            # The final list of messages
            prompt_messages = []
            # Looping the messages built in build_prompt
            # They only contain the template
            for templated_index, templated_message in enumerate(
                self.current_prompt.messages
            ):
                # If a message has a placeholder size, we need to replace it
                # With the N following messages, where N is the placeholder size
                if templated_message.placeholder_size:
                    for _ in range(templated_message.placeholder_size):
                        formatted_message = formatted_messages[
                            templated_index + placeholder_offset
                        ]
                        prompt_messages += [convert_message(formatted_message)]
                        # Increment the placeholder offset
                        placeholder_offset += 1
                    # Finally, decrement the placeholder offset by one
                    # Because the message representing the placeholder is now consumed
                    placeholder_offset -= 1
                # The current message is not a placeholder
                else:
                    formatted_message = formatted_messages[
                        templated_index + placeholder_offset
                    ]
                    # Update the role and formatted value, keep the template
                    prompt_messages += [
                        convert_message(
                            formatted_message, template=templated_message.template
                        )
                    ]
            # If the chat llm has more message than the initial chain prompt, append them
            # Typically happens with function agents
            if len(formatted_messages) > len(prompt_messages):
                prompt_messages += [
                    convert_message(m)
                    for m in formatted_messages[len(prompt_messages) :]
                ]
            # Finally set the prompt messages
            self.current_prompt.messages = prompt_messages
        # Non chat mode
        elif self.current_prompt.template:
            unique_message = formatted_messages[0]
            prompt_message = convert_message(
                unique_message, template=self.current_prompt.template
            )
            self.current_prompt.messages = [prompt_message]
            self.current_prompt.template = None
    # No current prompt, create it (formatted only)
    elif len(self.prompt_sequence):
        prompt_messages = [convert_message(m) for m in formatted_messages]
        prompt = Prompt(
            messages=prompt_messages,
            provider=provider,
            settings=settings,
        )
        self.prompt_sequence[-1] = prompt


def _on_llm_start(
    self: BaseLangchainCallbackHandler,
    serialized: Dict[str, Any],
    prompts: List[str],
    **kwargs: Any,
) -> None:
    invocation_params = kwargs.get("invocation_params")
    provider, settings = get_llm_settings(invocation_params, serialized)

    if self.current_prompt:
        self.current_prompt.formatted = prompts[0]
        self.current_prompt.provider = provider
        self.current_prompt.settings = settings
    elif len(self.prompt_sequence):
        prompt = Prompt(
            formatted=prompts[0],
            provider=provider,
            settings=settings,
        )
        self.prompt_sequence[-1] = prompt


class LangchainCallbackHandler(BaseLangchainCallbackHandler, BaseCallbackHandler):
    def on_error(self, error, **_):
        if error := self.create_error(error):
            run_sync(error.send())
            self.pop_sequence()

    on_tool_error = on_error
    on_llm_error = on_error
    on_chain_error = on_error

    def send_token(self, token: str, final: bool = False):
        stream = self.final_stream if final else self.stream
        if stream:
            run_sync(stream.stream_token(token))
            self.has_streamed_final_answer = final

    def add_message(self, message: Message):
        if message.author in IGNORE_LIST:
            return

        if self.stream:
            run_sync(self.stream.send())
            self.end_stream()

        else:
            run_sync(message.send())

    # Callbacks for various events

    def on_llm_start(
        self, serialized: Dict[str, Any], prompts: List[str], **kwargs: Any
    ) -> None:
        _on_llm_start(self, serialized, prompts, **kwargs)

    def on_chat_model_start(
        self,
        serialized: Dict[str, Any],
        messages: List[List[BaseMessage]],
        **kwargs: Any,
    ) -> None:
        _on_chat_model_start(self, serialized, messages, **kwargs)

    def on_llm_new_token(self, token: str, **kwargs: Any) -> None:
        if not self.stream:
            self.start_stream()
        self.send_token(token)

        if not self.stream_final_answer:
            return

        self.append_to_last_tokens(token)

        if self.answer_reached:
            if not self.final_stream:
                self.final_stream = Message(author=config.ui.name, content="")
            self.send_token(token, final=True)
        else:
            self.answer_reached = self.check_if_answer_reached()

    def on_llm_end(self, response: LLMResult, **kwargs: Any) -> None:
        if response.llm_output is not None:
            if "token_usage" in response.llm_output:
                token_usage = response.llm_output["token_usage"]
                if "total_tokens" in token_usage:
                    run_sync(
                        context.emitter.update_token_count(token_usage["total_tokens"])
                    )
        if self.current_prompt:
            self.current_prompt.completion = response.generations[0][0].text
        if self.final_stream:
            run_sync(self.final_stream.send())

    def on_chain_start(
        self, serialized: Dict[str, Any], inputs: Dict[str, Any], **kwargs: Any
    ) -> None:
        prompt = build_prompt(serialized, inputs)
        self.prompt_sequence.append(prompt)
        message = self.create_message(author=serialized["id"][-1])
        self.add_in_sequence(message)
        self.add_message(message)

    def on_chain_end(self, outputs: Dict[str, Any], **kwargs: Any) -> None:
        output_key = list(outputs.keys())[0]
        if output_key:
            parent_id = self.get_last_message().parent_id
            message = self.create_message(
                outputs[output_key], self.current_prompt, parent_id=parent_id
            )
            self.add_message(message)

        if self.prompt_sequence:
            self.prompt_sequence.pop()
        self.pop_sequence()

    def on_tool_start(
        self, serialized: Dict[str, Any], input_str: str, **kwargs: Any
    ) -> None:
        message = self.create_message(author=serialized["name"])
        self.add_in_sequence(message)
        self.add_message(message)

    def on_tool_end(
        self,
        output: str,
        observation_prefix: Optional[str] = None,
        llm_prefix: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        parent_id = self.get_last_message().parent_id
        message = self.create_message(output, None, parent_id=parent_id)
        self.add_message(message)
        self.pop_sequence()

    def on_text(self, text: str, **kwargs: Any) -> None:
        pass

    def on_agent_action(self, action: AgentAction, **kwargs: Any) -> Any:
        pass

    def on_agent_finish(self, finish: AgentFinish, **kwargs: Any) -> None:
        """Run on agent end."""
        pass


class AsyncLangchainCallbackHandler(BaseLangchainCallbackHandler, AsyncCallbackHandler):
    async def on_error(self, error, **_):
        if error := self.create_error(error):
            await error.send()
            self.pop_sequence()

    on_tool_error = on_error
    on_llm_error = on_error
    on_chain_error = on_error

    async def send_token(self, token: str, final: bool = False):
        stream = self.final_stream if final else self.stream
        if stream:
            await stream.stream_token(token)
            self.has_streamed_final_answer = final

    async def add_message(self, message: Message):
        if message.author in IGNORE_LIST:
            return

        if self.stream:
            await self.stream.send()
            self.end_stream()

        else:
            await message.send()

    # Callbacks for various events

    async def on_llm_start(
        self, serialized: Dict[str, Any], prompts: List[str], **kwargs: Any
    ) -> None:
        _on_llm_start(self, serialized, prompts, **kwargs)

    async def on_chat_model_start(
        self,
        serialized: Dict[str, Any],
        messages: List[List[BaseMessage]],
        **kwargs: Any,
    ) -> None:
        _on_chat_model_start(self, serialized, messages, **kwargs)

    async def on_llm_new_token(self, token: str, **kwargs: Any) -> None:
        if not self.stream:
            self.start_stream()
        await self.send_token(token)

        if not self.stream_final_answer:
            return

        self.append_to_last_tokens(token)

        if self.answer_reached:
            if not self.final_stream:
                self.final_stream = Message(author=config.ui.name, content="")
            await self.send_token(token, final=True)
        else:
            self.answer_reached = self.check_if_answer_reached()

    async def on_llm_end(self, response: LLMResult, **kwargs: Any) -> None:
        if response.llm_output is not None:
            if "token_usage" in response.llm_output:
                token_usage = response.llm_output["token_usage"]
                if "total_tokens" in token_usage:
                    await context.emitter.update_token_count(
                        token_usage["total_tokens"]
                    )
        if self.current_prompt:
            self.current_prompt.completion = response.generations[0][0].text
        if self.final_stream:
            await self.final_stream.send()

    async def on_chain_start(
        self, serialized: Dict[str, Any], inputs: Dict[str, Any], **kwargs: Any
    ) -> None:
        prompt = build_prompt(serialized, inputs)
        self.prompt_sequence.append(prompt)
        message = self.create_message(author=serialized["id"][-1])
        self.add_in_sequence(message)
        await self.add_message(message)

    async def on_chain_end(self, outputs: Dict[str, Any], **kwargs: Any) -> None:
        output_key = list(outputs.keys())[0]
        if output_key:
            parent_id = self.get_last_message().parent_id
            message = self.create_message(
                outputs[output_key], prompt=self.current_prompt, parent_id=parent_id
            )
            await self.add_message(message)
        if self.prompt_sequence:
            self.prompt_sequence.pop()
        self.pop_sequence()

    async def on_tool_start(
        self, serialized: Dict[str, Any], input_str: str, **kwargs: Any
    ) -> None:
        message = self.create_message(author=serialized["name"])
        self.add_in_sequence(message)
        await self.add_message(message)

    async def on_tool_end(
        self,
        output: str,
        observation_prefix: Optional[str] = None,
        llm_prefix: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        parent_id = self.get_last_message().parent_id
        message = self.create_message(output, parent_id=parent_id)
        await self.add_message(message)
        self.pop_sequence()

    async def on_text(self, text: str, **kwargs: Any) -> None:
        pass

    async def on_agent_action(self, action: AgentAction, **kwargs: Any) -> Any:
        pass

    async def on_agent_finish(self, finish: AgentFinish, **kwargs: Any) -> None:
        """Run on agent end."""
        pass



from contextlib import contextmanager

from chainlit.input_widget import Select, Slider, Tags
from chainlit.playground.provider import BaseProvider
from fastapi import HTTPException
from fastapi.responses import StreamingResponse

openai_common_inputs = [
    Slider(
        id="temperature",
        label="Temperature",
        min=0.0,
        max=2.0,
        step=0.01,
        initial=0.9,
        tooltip="Higher values like 0.8 will make the output more random, while lower values like 0.2 will make it more focused and deterministic.",
    ),
    Slider(
        id="max_tokens",
        label="Max Tokens",
        min=0.0,
        max=8000,
        step=1,
        initial=256,
        tooltip="The maximum number of tokens to generate in the chat completion.",
    ),
    Slider(
        id="top_p",
        label="Top P",
        min=0.0,
        max=1.0,
        step=0.01,
        initial=1.0,
        tooltip="An alternative to sampling with temperature, called nucleus sampling, where the model considers the results of the tokens with top_p probability mass. So 0.1 means only the tokens comprising the top 10% probability mass are considered.",
    ),
    Slider(
        id="frequency_penalty",
        label="Frequency Penalty",
        min=-2.0,
        max=2.0,
        step=0.01,
        initial=0.0,
        tooltip="Number between -2.0 and 2.0. Positive values penalize new tokens based on their existing frequency in the text so far, decreasing the model's likelihood to repeat the same line verbatim.",
    ),
    Slider(
        id="presence_penalty",
        label="Presence Penalty",
        min=-2.0,
        max=2.0,
        step=0.01,
        initial=0.0,
        tooltip="Number between -2.0 and 2.0. Positive values penalize new tokens based on whether they appear in the text so far, increasing the model's likelihood to talk about new topics.",
    ),
    Tags(
        id="stop",
        label="Stop Sequences",
        initial=[],
        tooltip="Up to 4 sequences where the API will stop generating further tokens.",
    ),
]


@contextmanager
def handle_openai_error():
    import openai

    try:
        yield
    except openai.error.Timeout as e:
        raise HTTPException(
            status_code=408,
            detail=f"OpenAI API request timed out: {e}",
        )
    except openai.error.APIError as e:
        raise HTTPException(
            status_code=500,
            detail=f"OpenAI API returned an API Error: {e}",
        )
    except openai.error.APIConnectionError as e:
        raise HTTPException(
            status_code=503,
            detail=f"OpenAI API request failed to connect: {e}",
        )
    except openai.error.InvalidRequestError as e:
        raise HTTPException(
            status_code=400,
            detail=f"OpenAI API request was invalid: {e}",
        )
    except openai.error.AuthenticationError as e:
        raise HTTPException(
            status_code=401,
            detail=f"OpenAI API request was not authorized: {e}",
        )
    except openai.error.PermissionError as e:
        raise HTTPException(
            status_code=403,
            detail=f"OpenAI API request was not permitted: {e}",
        )
    except openai.error.RateLimitError as e:
        raise HTTPException(
            status_code=429,
            detail=f"OpenAI API request exceeded rate limit: {e}",
        )


class ChatOpenAIProvider(BaseProvider):
    def format_message(self, message, prompt):
        message = super().format_message(message, prompt)
        return message.to_openai()

    async def create_completion(self, request):
        await super().create_completion(request)
        import openai

        env_settings = self.validate_env(request=request)

        deployment_id = self.get_var(request, "OPENAI_API_DEPLOYMENT_ID")

        if deployment_id:
            env_settings["deployment_id"] = deployment_id

        llm_settings = request.prompt.settings

        self.require_settings(llm_settings)

        messages = self.create_prompt(request)

        if "stop" in llm_settings:
            stop = llm_settings["stop"]

            # OpenAI doesn't support an empty stop array, clear it
            if isinstance(stop, list) and len(stop) == 0:
                stop = None

            llm_settings["stop"] = stop

        llm_settings["stream"] = True

        with handle_openai_error():
            response = await openai.ChatCompletion.acreate(
                **env_settings,
                messages=messages,
                **llm_settings,
            )

        async def create_event_stream():
            async for stream_resp in response:
                token = stream_resp.choices[0]["delta"].get("content", "")
                yield token

        return StreamingResponse(create_event_stream())


class OpenAIProvider(BaseProvider):
    def message_to_string(self, message):
        return message.to_string()

    async def create_completion(self, request):
        await super().create_completion(request)
        import openai

        env_settings = self.validate_env(request=request)

        deployment_id = self.get_var(request, "OPENAI_API_DEPLOYMENT_ID")

        if deployment_id:
            env_settings["deployment_id"] = deployment_id

        llm_settings = request.prompt.settings

        self.require_settings(llm_settings)

        prompt = self.create_prompt(request)

        if "stop" in llm_settings:
            stop = llm_settings["stop"]

            # OpenAI doesn't support an empty stop array, clear it
            if isinstance(stop, list) and len(stop) == 0:
                stop = None

            llm_settings["stop"] = stop

        llm_settings["stream"] = True

        with handle_openai_error():
            response = await openai.Completion.acreate(
                **env_settings,
                prompt=prompt,
                **llm_settings,
            )

        async def create_event_stream():
            async for stream_resp in response:
                token = stream_resp.get("choices")[0].get("text")
                yield token

        return StreamingResponse(create_event_stream())


openai_env_vars = {"api_key": "OPENAI_API_KEY"}

azure_openai_env_vars = {
    "api_key": "OPENAI_API_KEY",
    "api_type": "OPENAI_API_TYPE",
    "api_base": "OPENAI_API_BASE",
    "api_version": "OPENAI_API_VERSION",
}

ChatOpenAI = ChatOpenAIProvider(
    id="openai-chat",
    env_vars=openai_env_vars,
    name="ChatOpenAI",
    inputs=[
        Select(
            id="model",
            label="Model",
            values=["gpt-3.5-turbo", "gpt-3.5-turbo-16k", "gpt-4", "gpt-4-32k"],
            initial_value="gpt-3.5-turbo",
        ),
        *openai_common_inputs,
    ],
    is_chat=True,
)


AzureChatOpenAI = ChatOpenAIProvider(
    id="azure-openai-chat",
    env_vars=azure_openai_env_vars,
    name="AzureChatOpenAI",
    inputs=openai_common_inputs,
    is_chat=True,
)

OpenAI = OpenAIProvider(
    id="openai",
    name="OpenAI",
    env_vars=openai_env_vars,
    inputs=[
        Select(
            id="model",
            label="Model",
            values=["text-davinci-003", "text-davinci-002"],
            initial_value="text-davinci-003",
        ),
        *openai_common_inputs,
    ],
    is_chat=False,
)

AzureOpenAI = OpenAIProvider(
    id="azure",
    name="AzureOpenAI",
    env_vars=azure_openai_env_vars,
    inputs=openai_common_inputs,
    is_chat=False,
)
